#!/usr/bin/env python
import unittest
import numpy as np
from plico.types.zernike_coefficients import ZernikeCoefficients

__version__ = "$Id: zernike_coefficients_test.py 168 2016-12-06 21:33:24Z lbusoni $"


class ZernikeCoefficientsTest(unittest.TestCase):

    def setUp(self):
        self._nz = 21

    def _createZ(self):
        self._z = ZernikeCoefficients(
            np.arange(self._nz, dtype=np.float32), 42)

    def testNumpy(self):
        c1 = np.arange(self._nz, dtype=np.float32)
        z = ZernikeCoefficients.fromNumpyArray(c1)
        c2 = z.toNumpyArray()
        self.assertTrue(np.array_equal(c1, c2))


    def testComparison(self):
        coeffs1= np.arange(self._nz, dtype=np.float32)
        coeffs2= coeffs1.copy()
        counter = 42
        z1 = ZernikeCoefficients.fromNumpyArray(coeffs1, counter)
        z2 = ZernikeCoefficients.fromNumpyArray(coeffs2, counter)
        self.assertEqual(z1, z2)
        z3 = ZernikeCoefficients.fromNumpyArray(coeffs1* 2, counter)
        self.assertNotEqual(z1, z3)


    def testReturnDictionary(self):
        coeffs= np.arange(3)* 0.1
        counter= 12
        z = ZernikeCoefficients.fromNumpyArray(coeffs, counter)
        d= z.toDictionary()
        self.assertTrue(np.array_equal(list(d.keys()), np.array([2, 3, 4])))
        self.assertTrue(np.array_equal(list(d.values()), np.array([0, 0.1, 0.2])))


    def testReturnLength(self):
        self._createZ()
        self.assertEqual(self._z.numberOfModes(), self._nz)


    def testGetZ(self):
        self._createZ()
        wantZernIndexes= [2, 4, 21]
        shouldGet= (np.array(wantZernIndexes)- 2).astype(np.float32)
        didGet= self._z.getZ(wantZernIndexes)
        self.assertTrue(np.array_equal(shouldGet, didGet),
                        "wanted %s got %s" % (shouldGet, didGet))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
import unittest
from plico.utils.calibration_installer import CalibrationInstaller
import tempfile
import os
import shutil

__version__ = "$Id: calibration_installer_test.py 26 2018-01-26 19:06:25Z lbusoni $"


class CalibrationInstallerTest(unittest.TestCase):


    def setUp(self):
        self._srcDir= tempfile.mkdtemp()
        self._destDir= tempfile.mkdtemp()
        print('src dir %s' % self._srcDir)
        print('dest dir %s' % self._destDir)
        self._createSourceFolder()


    def _createSourceFolder(self):
        self._createFile(os.path.join(self._srcDir, 'root1.txt'), "root1")
        self._fooDir= os.path.join(self._srcDir, 'foo')
        os.makedirs(self._fooDir)
        self._createFile(os.path.join(self._fooDir, 'foo1.txt'), "foo1")


    def _createFile(self, filename, text):
        with open(filename, "w") as out:
            out.write(text)


    def tearDown(self):
        shutil.rmtree(self._srcDir)
        shutil.rmtree(self._destDir)



    def testHappyPathWithForcedSrcAndDest(self):
        appDirs= None
        packageName= 'whatever'
        calibInstaller= CalibrationInstaller(
            appDirs,
            packageName,
            src=self._srcDir,
            dest=os.path.join(self._destDir, 'copyhere'))
        self.assertFalse(calibInstaller.doesCalibrationFolderExists())
        calibInstaller.installCalibrationFilesFromPackage()
        self.assertTrue(calibInstaller.doesCalibrationFolderExists())



if __name__ == "__main__":
    unittest.main()



import os
import shutil


__version__= "$Id: addtree.py 50 2018-04-26 21:42:20Z lbusoni $"


def mkdirp(dirname):
    if os.path.isdir(dirname):
        return
    os.makedirs(dirname)


def addtree(src, dst, symlinks=False):
    names = os.listdir(src)
    mkdirp(dst)
    errors = []
    for name in names:
        srcname = os.path.join(src, name)
        dstname = os.path.join(dst, name)
        try:
            if symlinks and os.path.islink(srcname):
                linkto = os.readlink(srcname)
                os.symlink(linkto, dstname)
            elif os.path.isdir(srcname):
                addtree(srcname, dstname, symlinks)
            else:
                shutil.copy2(srcname, dstname)
            # XXX What about devices, sockets etc.?
        except OSError as why:
            errors.append((srcname, dstname, str(why)))
        # catch the Error from the recursive copytree so that we can
        # continue with other files
        except shutil.Error as err:
            errors.extend(err.args[0])
    try:
        shutil.copystat(src, dst)
    except OSError as why:
        # can't copy file access times on Windows
        if why.winerror is None:
            errors.extend((src, dst, str(why)))
    if errors:
        raise shutil.Error(errors)



from plico.utils.logger import Logger


class Hackerable(object):

    def __init__(self, logger=None):
        self._hackerableExecuteCnt= 0
        self._hackerableEvalCnt= 0
        if logger is None:
            self._logger= Logger.of('Hackerable')


    def eval(self, expression):
        self._logger.warn("Evaluating '%s' (counter: %d)" % (
            expression, self._hackerableEvalCnt))
        self._hackerableEvalCnt+= 1
        return repr(eval(expression))


    def execute(self, statement):
        self._logger.warn("Executing '%s' (counter: %d)" % (
            statement, self._hackerableExecuteCnt))
        self._hackerableExecuteCnt+= 1
        exec(statement)



import time
from plico.types.server_info import ServerInfo
from plico.utils.logger import Logger


__version__= "$Id: deformable_mirror_controller.py 26 2018-01-26 19:06:25Z lbusoni $"


class ServerInfoable(object):

    def __init__(self, name, ports, logger=None):
        self._serverName= name
        self._creationTime= time.time()
        self._ports= ports
        if logger is None:
            self._logger= Logger.of('ServerInfoable')


    def serverInfo(self):
        return ServerInfo(
            self._serverName,
            time.time() - self._creationTime,
            self._ports.SERVER_HOSTNAME,
            self._ports.SERVER_REPLY_PORT)



import os

__version__= "$Id: fits_file_based_calibration_manager.py 25 2018-01-26 19:00:40Z lbusoni $"


class FitsFileBasedCalibrationManager(object):
    def _createFoldersIfMissing(self, filename):
        if not os.path.isdir(os.path.dirname(filename)):
            os.makedirs(os.path.dirname(filename))



#!/usr/bin/env python
import unittest
import os
from plico.utils.starter_script_creator_base import StarterScriptCreatorBase
import shutil

TEST_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                        "./tmp/")


class MyStarterScriptCreator(StarterScriptCreatorBase):

    def __init__(self):
        StarterScriptCreatorBase.__init__(self)
        self.destination_path = os.path.join(
            TEST_DIR, 'destination', 'path')
        self.executable = os.path.join('executable', 'module.py')
        self.config_section = 'my_config_section'

    def installExecutables(self):
        self._createAStarterScript(
            self.destination_path,
            self.executable,
            self.config_section
        )


class StarterScriptCreatorBaseTest(unittest.TestCase):

    LOG_DIR = os.path.join(TEST_DIR, "log")
    CONF_FILE = os.path.join("path", "to", "config", "file.conf")
    BIN_DIR = os.path.join(TEST_DIR, "apps", "bin")
    SOURCE_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                              "../..")

    def setUp(self):
        self.ssc = MyStarterScriptCreator()
        self.ssc.setInstallationBinDir(self.BIN_DIR)
        self.ssc.setPythonPath(self.SOURCE_DIR)
        self.ssc.setConfigFileDestination(self.CONF_FILE)
        self.ssc.installExecutables()

    def tearDown(self):
        if os.path.exists(TEST_DIR):
            shutil.rmtree(TEST_DIR)

    def _is_file(self, path):
        import sys
        if sys.version_info.major == 2:
            import os.path
            return os.path.isfile(path)
        elif sys.version_info.major == 3:
            import pathlib
            return pathlib.Path(path).resolve().is_file()
        else:
            raise Exception("unknown python version")

    def test_create_script(self):

        want_file_in = self.ssc.destination_path
        self.assertTrue(self._is_file(want_file_in))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import os
import re
import configparser
import appdirs
from pkg_resources import resource_filename
from plico.utils.calibration_installer import CalibrationInstaller
from plico.utils.addtree import mkdirp


class LogFolderManager():

    def __init__(self, appDirs):
        self._appdirs = appDirs

    def getLogFolderPath(self):
        return self._appdirs.user_log_dir

    def doesLogFolderExists(self):
        return os.path.isdir(self.getLogFolderPath())

    def createLogFolder(self):
        mkdirp(os.path.dirname(self.getLogFolderPath()))


class Configuration(object):

    def __init__(self):
        NOT_INITIALIZED = None
        self._cfg = configparser.ConfigParser()

    def networkHostName(self, section):
        return self.getValue(section, 'host')

    def calibrationRootDir(self):
        return self._getCalibDir()

    def loggingDir(self):
        return self._getLogDir()

    def deviceModel(self, section):
        return self.getValue(section, 'model')

    def deviceName(self, section):
        return self.getValue(section, 'name')

    def basePort(self, section):
        return self.getValue(section, 'port', getint=True)

    def logLevel(self, section):
        return self.getValue(section, 'log_level')

    def numberedSectionList(self, prefix=''):
        pattern = prefix + '\\d+'
        return list(filter(lambda x: re.search(pattern, x), self._cfg.sections()))

    def _assertSectionExists(self, section):
        if section not in self._cfg.sections():
            raise Exception('Section "%s" not found in configuration file' %
                            section)

    def getValue(self, section, entry, getint=False,
                 getfloat=False, getboolean=False):
        if entry not in list(dict(self._cfg.items(section)).keys()):
            raise KeyError(
                'Section "%s" must contain a "%s" entry. File is %s. Keys %s' %
                (section, entry, self._filename,
                 list(dict(self._cfg.items(section)).keys())))
        if getint:
            try:
                value = self._cfg.getint(section, entry)
            except ValueError:
                raise ValueError(
                    'Value for entry "%s" in section "%s"'
                    ' must be an integer number' % (entry, section))
        elif getfloat:
            try:
                value = self._cfg.getfloat(section, entry)
            except ValueError:
                raise ValueError(
                    'Value for entry "%s" in section "%s"'
                    ' must be a float number' % (entry, section))
        elif getboolean:
            try:
                value = self._cfg.getboolean(section, entry)
            except ValueError:
                raise ValueError(
                    'Value for entry "%s" in section "%s"'
                    ' must be a boolean' % (entry, section))
        else:
            value = self._cfg.get(section, entry)
        return value

    def load(self, filename):
        if not os.path.exists(filename):
            raise Exception('Configuration file not found: %s' % filename)
        self._cfg.read(filename)
        self._filename = filename

    def _getLogDir(self):
        try:
            return self.getValue('global', 'force_log_dir')
        except Exception:
            lfm = LogFolderManager(self._getAppDirs())
            lfm.createLogFolder()
            return lfm.getLogFolderPath()

    def _getCalibDir(self):
        try:
            return self.getValue('global', 'force_calib_folder_dest')
        except Exception:
            calibInst = CalibrationInstaller(
                self._getAppDirs(), self._getPythonPackageName())
            calibInst.installCalibrationFilesFromPackage()
            return calibInst.getCalibrationFolderPath()

    def _getPythonPackageName(self):
        return self.getValue('global', 'python_package_name')

    def _getAppName(self):
        return self.getValue('global', 'app_name')

    def _getAppAuthor(self):
        return self.getValue('global', 'app_author')

    def _getAppDirs(self):
        return appdirs.AppDirs(self._getAppName(),
                               self._getAppAuthor())



import abc
from plico.utils.decorator import returns
from six import with_metaclass


__version__= "$Id: stepable.py 25 2018-01-26 19:00:40Z lbusoni $"


class Stepable(with_metaclass(abc.ABCMeta, object)):


    @abc.abstractmethod
    def step(self):
        assert False


    @abc.abstractmethod
    @returns(bool)
    def isTerminated(self):
        assert False



#!/usr/bin/env python

import unittest
import numpy as np
from plico.utils.zernike_generator import ZernikeGenerator

__version__ = "$Id: zernike_generator_test.py 26 2018-01-26 19:06:25Z lbusoni $"


class TestZernikeGenerator(unittest.TestCase):


    def setUp(self):
        self._nPixels= 10
        self.generator= ZernikeGenerator(self._nPixels)


    def testPiston(self):
        piston= self.generator[1]
        self.assertEqual(piston.shape, (self._nPixels, self._nPixels))
        self.assertTrue(piston.mask[0, 0] == True)
        self.assertAlmostEqual(piston[self._nPixels // 2, self._nPixels // 2],
                               1.)


    def testTilt(self):
        nPx=128
        generator= ZernikeGenerator(nPx)
        tilt= generator[3]
        self.assertAlmostEqual(tilt[-1, nPx// 2], 2.* (1 - 1. / nPx))


    def testOdd(self):
        nPx=137
        generator= ZernikeGenerator(nPx)
        tilt= generator[17]
        self.assertAlmostEqual(tilt[int(nPx / 2.),
                                    int(nPx / 2.)], 0.)


    def testDegree(self):
        m, n= self.generator._degree(1)
        self.assertEqual(m, 0)
        self.assertEqual(n, 0)
        m, n= self.generator._degree(2)
        self.assertEqual(m, 1)
        self.assertEqual(n, 1)
        m, n= self.generator._degree(3)
        self.assertEqual(m, 1)
        self.assertEqual(n, 1)
        m, n= self.generator._degree(4)
        self.assertEqual(m, 2)
        self.assertEqual(n, 0)
        m, n= self.generator._degree(2000000)
        self.assertEqual(m, 1999)
        self.assertEqual(n, 999)


    def testRmn(self):
        self.assertTrue(np.allclose(
            self.generator._rnm(0, 0, np.array([0, 0.5, 1])),
            np.array([1, 1, 1])))
        self.assertTrue(np.allclose(
            self.generator._rnm(1, 1, np.array([0, 0.5, 1])),
            np.array([0, 0.5, 1])))
        self.assertTrue(np.allclose(
            self.generator._rnm(2, 0, np.array([0, 0.5, 1])),
            np.array([-1, -0.5, 1])))
        self.assertTrue(np.allclose(
            self.generator._rnm(5, 1, np.array([0, 0.5, 1])),
            np.array([0, 0.3125, 1])))


    def testPolar(self):
        self.assertTrue(np.allclose(
            self.generator._polar(1, np.array([0, 0.5, 1]),
                                  np.array([0, 1, 5]) * np.pi / 4),
            np.array([1, 1, 1])))
        self.assertTrue(np.allclose(
            self.generator._polar(2, np.array([0, 0.5, 1]),
                                  np.array([0, 1, 5]) * np.pi / 3),
            np.array([0, 0.5, 1])))
        self.assertTrue(np.allclose(
            self.generator._polar(3, np.array([0, 0.5, 1]),
                                  np.array([0, 1, 5]) * np.pi / 3),
            np.array([0, 0.866025, -1.73205])))
        self.assertTrue(np.allclose(
            self.generator._polar(16, np.array([0, 0.5, 1]),
                                  np.array([0, 1, 5]) * np.pi / 3),
            np.array([0, 0.541266, 1.73205])))


    def testDerivativeX(self):
        gammaX= self.generator._derivativeCoeffX(30)
        self.assertAlmostEqual(gammaX[1, 0], 2)
        self.assertAlmostEqual(gammaX[15, 3], 6)
        self.assertAlmostEqual(gammaX[16, 4], 3 * np.sqrt(2))


    def testDerivativeX2(self):
        gammaX= self.generator._derivativeCoeffX(2)
        self.assertTrue(np.allclose(gammaX,
                        np.array([[0., 0.], [2., 0.]])),
                        "%s" % str(gammaX))


    def testDerivativeY3(self):
        gammaY= self.generator._derivativeCoeffY(3)
        self.assertTrue(np.allclose(gammaY,
                        np.array([[0., 0., 0.],
                                  [0., 0., 0.],
                                  [2., 0., 0.]])),
                        "%s" % str(gammaY))


    def testDerivativeY(self):
        gammaY= self.generator._derivativeCoeffY(30)
        self.assertAlmostEqual(gammaY[2, 0], 2)
        self.assertAlmostEqual(gammaY[16, 3], 6)
        self.assertAlmostEqual(gammaY[9, 4], -2 * np.sqrt(3))



    def testRmsIsOne(self):
        zg= ZernikeGenerator(512)
        self.assertAlmostEqual(1., np.std(zg.getZernike(4)), 4)


    def testGetItem(self):
        a= self.generator[2]
        b= self.generator.getZernike(2)
        self.assertTrue(np.allclose(a, b))


    def testGetZernikeDict(self):
        dd= self.generator.getZernikeDict([1, 2, 10])
        self.assertTrue(np.allclose([1, 2, 10], list(dd.keys())))
        self.assertTrue(np.allclose(dd[10], self.generator.getZernike(10)))


    def testGetDerivativeOfTip(self):
        got= self.generator.getDerivativeX(2)
        wanted= 2* np.ones((self._nPixels, self._nPixels))
        self.assertTrue(np.allclose(wanted, got),
                        "got %s, wanted %s" % (str(got), str(wanted)))
        got= self.generator.getDerivativeY(2)
        wanted= 0.* np.ones((self._nPixels, self._nPixels))
        self.assertTrue(np.allclose(wanted, got),
                        "got %s, wanted %s" % (str(got), str(wanted)))


    def testGetDerivativeOfTilt(self):
        got= self.generator.getDerivativeX(3)
        wanted= 0.* np.ones((self._nPixels, self._nPixels))
        self.assertTrue(np.allclose(wanted, got),
                        "got %s, wanted %s" % (str(got), str(wanted)))
        got= self.generator.getDerivativeY(3)
        wanted= 2.* np.ones((self._nPixels, self._nPixels))
        self.assertTrue(np.allclose(wanted, got),
                        "got %s, wanted %s" % (str(got), str(wanted)))


#     def testGetDerivativeOfFocus(self):
#         got= self.generator.getDerivativeX(4)
#         wanted= np.(self._nPixels)
#         self.assertTrue(np.allclose(wanted, got),
#                         "got %s, wanted %s" % (str(got), str(wanted)))



if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
import unittest
import tempfile
import shutil
import os
from plico.utils.addtree import addtree

__version__ = "$Id: addtree_test.py 26 2018-01-26 19:06:25Z lbusoni $"


class Test(unittest.TestCase):


    def setUp(self):
        self._srcDir= tempfile.mkdtemp()
        self._destDir= tempfile.mkdtemp()
        print('src dir %s' % self._srcDir)
        print('dest dir %s' % self._destDir)
        self._createSourceFolder()


    def _createSourceFolder(self):
        self._createFile(os.path.join(self._srcDir, 'root1.txt'), "root1")
        self._fooDir= os.path.join(self._srcDir, 'foo')
        os.makedirs(self._fooDir)
        self._createFile(os.path.join(self._fooDir, 'foo1.txt'), "foo1")
        self._createFile(os.path.join(self._fooDir, 'foo2.txt'), "foo2")
        self._booDir= os.path.join(self._srcDir, 'boo')
        os.makedirs(self._booDir)
        self._createFile(os.path.join(self._booDir, 'boo1.txt'), "boo1")


    def _createFile(self, filename, text):
        with open(filename, "w") as out:
            out.write(text)


    def tearDown(self):
        print(self._listSrcDir())
        print(self._listDestDir())
        shutil.rmtree(self._srcDir)
        shutil.rmtree(self._destDir)


    def testHappyPath(self):
        addtree(self._srcDir, self._destDir)
        self._checkAllCopied()


    def _checkAllCopied(self):
        self._checkDestDir('foo')
        self._checkDestDir('boo')
        self._checkDestFile(os.path.join('boo', 'boo1.txt'))
        self._checkDestFile(os.path.join('foo', 'foo1.txt'))
        self._checkDestFile(os.path.join('foo', 'foo2.txt'))
        self._checkDestFile('root1.txt')


    def testCopyOnEmptyDestAndCopyAgain(self):
        addtree(self._srcDir, self._destDir)
        addtree(self._srcDir, self._destDir)
        self._checkAllCopied()


    def testNewlyAddedFilesAreNotRemovedWhenCopying(self):
        addtree(self._srcDir, self._destDir)

        self._createFile(os.path.join(self._destDir, 'root2.txt'), "root2")
        newdir= os.path.join(self._destDir, 'new')
        os.makedirs(newdir)
        self._createFile(os.path.join(newdir, 'new1.txt'), "new1")

        addtree(self._srcDir, self._destDir)
        self._checkAllCopied()
        self._checkNewOnesAreStillThere()


    def testModifiedFilesAreNotOverwrittenWhenCopying(self):
        addtree(self._srcDir, self._destDir)

        fileToModify= os.path.join(self._srcDir, 'root1.txt')
        self._modifyText(fileToModify, "new text for root1")

        addtree(self._srcDir, self._destDir)
        self._checkAllCopied()
        self.assertEqual("new text for root1",
                         self._readFile(fileToModify))


    def _readFile(self, filePath):
        with open(filePath) as file:
            data = file.read()
        return data

    def _modifyText(self, file, newText):
        os.remove(file)
        self._createFile(file, newText)


    def _checkNewOnesAreStillThere(self):
        self._checkDestFile('root2.txt')
        self._checkDestDir('new')
        self._checkDestFile(os.path.join('new', 'new1.txt'))


    def _checkDestDir(self, dirName):
        dirPath= os.path.join(self._destDir, dirName)
        self.assertTrue(os.path.isdir(dirPath),
                        "%s not found" % dirPath)


    def _checkDestFile(self, fileName):
        filePath= os.path.join(self._destDir, fileName)
        self.assertTrue(os.path.isfile(filePath),
                        "%s not found" % filePath)


    def _listSrcDir(self):
        return "src has: %s" % str([x for x in os.walk(self._srcDir)])


    def _listDestDir(self):
        return "dest has: %s" % str([x for x in os.walk(self._destDir)])


if __name__ == "__main__":
    unittest.main()



import subprocess
import logging


__version__= "$Id: kill_process_by_name.py 25 2018-01-26 19:00:40Z lbusoni $"


def killProcessByName(processName):
    proc= subprocess.Popen("pgrep -f %s" % processName,
                           stdout=subprocess.PIPE,
                           stderr=subprocess.PIPE,
                           shell=True)

    pgrepPID= proc.pid
    pids= []
    for each in proc.stdout:
        if int(each) == pgrepPID:
            logging.debug("Skipping pgrep proc with PID %d" % (
                pgrepPID))
            continue
        logging.debug("%s with PID %s" % (processName, each))
        pids.append(int(each))

    logging.info("number of processes '%s': %d" % (processName, len(pids)))
    for each in pids:
        cmd= "kill -KILL %d" % each
        logging.debug("Executing %s" % cmd)
        exitCode= subprocess.call(cmd, shell=True)
        assert exitCode == 0, "Terminating %s with PID %d" % (
            processName, each)



import unittest
from plico.utils.control_loop import FaultTolerantControlLoop, IntolerantControlLoop
from plico.utils.logger import DummyLogger, DummyLoggerFactory
from plico.utils.stepable import Stepable
from test.fake_time_mod import FakeTimeMod

class MyException(Exception):
    pass

class RaisingStepable(Stepable):

    def __init__(self, terminate_after_iteration=3):
        self.iter = 0
        self.max_iter = terminate_after_iteration
        
    def step(self):
        self.iter += 1
        raise MyException()
    
    def isTerminated(self):
        return self.iter >= self.max_iter

class IntolerantControlLoopTest(unittest.TestCase):

    def test_intolerant_loop_always_raises(self):
        raising_stepable = RaisingStepable()
        logger = DummyLoggerFactory().getLogger('test_logger')
        loop = IntolerantControlLoop(raising_stepable, logger, timeModule=FakeTimeMod())
        self.assertRaises(MyException, loop.start)

class FaultTolerantControlLoopTest(unittest.TestCase):

    def test_fault_tolerant_loop_never_raises(self):
        raising_stepable = RaisingStepable()
        logger = DummyLoggerFactory().getLogger('test_logger')
        loop = FaultTolerantControlLoop(raising_stepable, logger, timeModule=FakeTimeMod())
        loop.start()
        self.assertEqual(raising_stepable.iter, raising_stepable.max_iter)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import unittest
from plico.utils.barrier import Predicate, Barrier, BarrierTimeout,\
    FunctionPredicate
from plico.utils.decorator import override


__version__="$Id: barrier_test.py 26 2018-01-26 19:06:25Z lbusoni $"


class Dummy:

    def __init__(self):
        self._isTrueInvocations= 0


    def isTrue(self):
        if self._isTrueInvocations < 3:
            self._isTrueInvocations+= 1
            return False
        return True


    def getIsTrueInvocations(self):
        return self._isTrueInvocations


class MyTimeModule():
    def __init__(self):
        self._count= 0
        self._returnValues= None

    def setReturnValues(self, returnValues):
        self._returneValues= returnValues
        self._count= 0

    def time(self):
        ret= self._returnValues[self._count]
        self._count+= 1
        return ret


class MyPredicate(Predicate):

    def __init__(self):
        self._errorMessage= None
        self._isFullfilledValues= [False]
        self._count= 0

    def setErrorMessage(self, msg):
        self._errorMessage= msg

    @override
    def errorMessage(self):
        return self._errorMessage

    def setFullfilledValues(self, fullfilledValues):
        self._isFullfilledValues= fullfilledValues
        self._count= 0

    @override
    def isFullfilled(self):
        nElements= len(self._isFullfilledValues)
        ret= self._isFullfilledValues[self._count % nElements]
        self._count+= 1
        return ret



class BarrierStaticMethodTest(unittest.TestCase):

    def trueFunction(self):
        return True


    def falseFunction(self):
        return False


    def test_should_return_True(self):
        Barrier.waitUntil(self.trueFunction, 0.5, 0.15)


    def test_should_raise_timeout_exception(self):
        TIMEOUT_IN_SEC= 0.5
        timeModule= MyTimeModule()
        timeModule.setReturnValues([0, TIMEOUT_IN_SEC])
        self.assertRaises(Exception, Barrier.waitUntil,
                          self.falseFunction,
                          TIMEOUT_IN_SEC, 0.15, timeModule)


    def test_invocations(self):
        dummy= Dummy()

        Barrier.waitUntil(dummy.isTrue, 2, 0.01)
        self.assertTrue(dummy.getIsTrueInvocations() == 3)



class BarrierDynamicMethodTest(unittest.TestCase):

    def setUp(self):
        class TimeFakeModule(object):

            def __init__(self):
                self._currentTime= 0


            def time(self):
                self._currentTime+= 0.1
                return self._currentTime


            def sleep(self, sleepDurationInSec):
                pass

        self.predicate= MyPredicate()
        self.predicate.setErrorMessage("predicate failure")

        self.barrier= Barrier(
            timeoutSec=3, pollingPeriodSec=0.1,
            timeModule=TimeFakeModule())


    def test_should_wait_for_predicate(self):
        self.predicate= MyPredicate()
        self.predicate.setFullfilledValues([False, True])
        self.barrier.waitFor(self.predicate)


    def test_should_detect_timeout(self):
        self.predicate= MyPredicate()
        self.predicate.setFullfilledValues([False])
        self.predicate.setErrorMessage("Tux failure")

        exceptionThrown= False
        try:
            self.barrier.waitFor(self.predicate)
        except BarrierTimeout as e:
            exceptionThrown= True
            self.assertTrue("Tux failure" in str(e))
        self.assertTrue(exceptionThrown)


    def test_should_forward_predicate_failures(self):
        class FailingPredicate(Predicate):

            @override
            def isFullfilled(self):
                raise Exception("test")


            @override
            def errorMessage(self):
                assert False

        try:
            self.barrier.waitFor(FailingPredicate())
        except BarrierTimeout:
            self.fail("No barrier timeout expected")
        except Exception as e:
            self.assertEqual("test", str(e))
            return

        self.fail("Exception expected")


class FunctionPredicateTest(unittest.TestCase):


    def test_creation(self):
        def foo(a, b):
            return a == b
        pred= FunctionPredicate.create(foo, 3, 4)
        self.assertFalse(pred.isFullfilled())

        pred= FunctionPredicate.create(foo, 1, 1)
        self.assertTrue(pred.isFullfilled())

        print("pred:", pred.errorMessage())
        self.assertTrue("foo" in pred.errorMessage())


if __name__ == "__main__":
    unittest.main()



import numpy as np

__version__= "$Id: zernike_coefficients.py 35 2018-01-28 13:33:16Z lbusoni $"


class ZernikeCoefficients(object):
    FIRST_ZERNIKE_MODE= 2

    def __init__(self, coefficients, counter=0):
        self._coefficients= coefficients
        self._counter= counter

    def zernikeIndexes(self):
        return np.arange(self.FIRST_ZERNIKE_MODE,
                         self.FIRST_ZERNIKE_MODE+ self.numberOfModes())


    def numberOfModes(self):
        return len(self._coefficients)


    def getZ(self, zernikeIndexes):
        return self.toNumpyArray()[np.array(zernikeIndexes)-
                                   self.FIRST_ZERNIKE_MODE]


    def toDictionary(self):
        keys= self.zernikeIndexes()
        values= self._coefficients
        return dict(list(zip(keys, values)))


    def toNumpyArray(self):
        return self._coefficients


    @staticmethod
    def fromNumpyArray(coefficientsAsNumpyArray, counter=0):
        return ZernikeCoefficients(np.array(coefficientsAsNumpyArray), counter)


    def counter(self):
        return self._counter


    def setCounter(self, counter):
        self._counter = counter


    def __eq__(self, o):
        if self._counter != o._counter:
            return False
        if not np.array_equal(self._coefficients, o._coefficients):
            return False
        return True


    def __ne__(self, o):
        return not self.__eq__(o)


    def __str__(self):
        return str(self._coefficients)




import abc
from plico.utils.decorator import returns
from six import with_metaclass


class Convergeable(with_metaclass(abc.ABCMeta, object)):

    @abc.abstractmethod
    @returns(bool)
    def hasConverged(self):
        assert False

    @abc.abstractmethod
    def performOneConvergenceStep(self):
        assert False

    @abc.abstractmethod
    def measureConvergence(self):
        assert False



import logging
import os
import datetime
import sys
import subprocess


class StarterScriptCreatorBase(object):

    def __init__(self):
        NOT_INITIALIZED = None
        self._binDir = NOT_INITIALIZED
        self._pythonPath = NOT_INITIALIZED
        self._configFileDest = NOT_INITIALIZED

    def _now(self):
        return datetime.datetime.now()

    def _createFoldersIfMissing(self, filename):
        if not os.path.isdir(os.path.dirname(filename)):
            os.makedirs(os.path.dirname(filename))

    def setInstallationBinDir(self, binDir):
        self._binDir = binDir

    def setPythonPath(self, pythonPath):
        self._pythonPath = pythonPath

    def setConfigFileDestination(self, configFileDest):
        self._configFileDest = configFileDest

    def _createAStarterScript(self,
                              destinationPath,
                              executableSrc,
                              configFileSection,
                              command='python'):
        logging.info("Creating starter script %s" % destinationPath)
        self._createFoldersIfMissing(destinationPath)
        with open(destinationPath, "w") as text_file:
            text_file.write("#!/bin/bash\n")
            text_file.write("# Auto-generated on %s \n\n" % (
                str(self._now())))
            text_file.write('export PYTHONPATH="%s":$PYTHONPATH\n' %
                            self._pythonPath)
            text_file.write("%s '%s' \"%s\" %s\n" % (
                command,
                os.path.join(self._pythonPath, executableSrc),
                self._configFileDest,
                configFileSection))

        self._setExecutableFlag(destinationPath)

    def _setExecutableFlag(self, filePath):
        if sys.platform == "win32":
            return
        self._execute('chmod +x "%s"' % filePath)

    def _execute(self, cmd):
        logging.info(cmd)
        exitCode = subprocess.call(cmd, shell=True)
        assert 0 == exitCode, "Command '%s' must succeed" % cmd

    def installExecutables(self):
        assert False, 'method to be implemented by derived classes'



from __future__ import print_function

import logging.handlers
import os
import sys
import signal
from plico.utils.configuration import Configuration
from plico.rpc.zmq_remote_procedure_call import ZmqRemoteProcedureCall
from plico.rpc.sockets import Sockets
from plico.rpc.zmq_ports import ZmqPorts


class BaseRunner(object):

    def __init__(self):
        INITIALIZED_LATER = None
        self._configFilePath = INITIALIZED_LATER
        self._rpc = INITIALIZED_LATER
        self._configuration = INITIALIZED_LATER
        self._sockets = INITIALIZED_LATER
        self._argv = INITIALIZED_LATER

    def _logRunning(self):
        self._logger.notice(self.RUNNING_MESSAGE)
        sys.stdout.flush()

    def _createFoldersIfMissing(self, filename):
        if not os.path.isdir(os.path.dirname(filename)):
            os.makedirs(os.path.dirname(filename))

    def _determineLoggingLevel(self):
        try:
            levelAsStr = self._configuration.logLevel(
                self._configurationSection)
        except KeyError:
            levelAsStr = 'debug'
        if levelAsStr == 'fatal':
            loggingLevel = logging.CRITICAL
        elif levelAsStr == 'error':
            loggingLevel = logging.ERROR
        elif levelAsStr == 'warning':
            loggingLevel = logging.WARNING
        elif levelAsStr == 'info':
            loggingLevel = logging.INFO
        elif levelAsStr == 'debug':
            loggingLevel = logging.DEBUG
        else:
            loggingLevel = logging.WARNING
        return loggingLevel

    def _setUpLogging(self, logFilePath):
        loggingLevel = self._determineLoggingLevel()
        print(('Setting up logging in file %s level %s' % (
            logFilePath, loggingLevel)))
        self._createFoldersIfMissing(logFilePath)
        FORMAT = '%(asctime)s %(levelname)s %(message)s'
        f = logging.Formatter(fmt=FORMAT)
        rotHandler = logging.handlers.RotatingFileHandler(
            logFilePath, encoding='utf8',
            maxBytes=10000000, backupCount=10)
        handlers = [rotHandler]
        root_logger = logging.getLogger()
        root_logger.setLevel(loggingLevel)
        for h in handlers:
            h.setFormatter(f)
            h.setLevel(loggingLevel)
            root_logger.addHandler(h)
        rotHandler.doRollover()

    def _createConfiguration(self):
        self._configuration = Configuration()
        self._configuration.load(self._configFilePath)

    def _createZmqBasedRPC(self):
        zmqPorts = ZmqPorts.fromConfiguration(self._configuration,
                                              self._configurationSection)
        self._rpc = ZmqRemoteProcedureCall()
        self._sockets = Sockets(zmqPorts, self._rpc)

    def _checkCommandLine(self, argv):
        if 3 > len(argv):
            programName = argv[0]
            print("Usage:", file=sys.stderr)
            print("\t%s <configuration_file> <configuration_section>" %
                  programName, file=sys.stderr)
            print("Example:", file=sys.stderr)
            print(("\t%s /home/plico/.plico/plico.conf") %
                  programName, file=sys.stderr)
            print("\nargv was: %s" % str(argv))
            sys.exit(os.EX_USAGE)
        print('argv: %s' % str(argv))

    def _parseCommandLine(self, argv):
        self._checkCommandLine(argv)
        self._configFilePath = argv[1]
        self._configurationSection = argv[2]

    def _logFilePath(self):
        return os.path.join(self._configuration.loggingDir(),
                            '%s.log' % self._configurationSection)

    def start(self, argv):
        self._argv = argv
        self._parseCommandLine(argv)
        self._createConfiguration()
        self._setUpLogging(self._logFilePath())
        self._createZmqBasedRPC()
        self._registerHandlers()
        return self.run()

    def run(self):
        pass

    def getConfigFilePath(self):
        return self._configFilePath

    @property
    def name(self):
        return self._configuration.getValue(
            self._configurationSection,
            'name')

    @property
    def configuration(self):
        return self._configuration

    def getConfigurationSection(self):
        return self._configurationSection

    def rpc(self):
        return self._rpc

    def sockets(self):
        return self._sockets

    def argv(self):
        return self._argv

    def terminate(self, signal, frame):
        pass

    def _registerHandlerForSigInt(self):
        signal.signal(signal.SIGINT, self.terminate)

    def _registerHandlerForSigTerm(self):
        signal.signal(signal.SIGTERM, self.terminate)

    def _registerHandlerForSigKill(self):
        try:
            signal.signal(signal.SIGKILL, self.terminate)
        except Exception:
            pass

    def _registerHandlers(self):
        self._registerHandlerForSigInt()
        self._registerHandlerForSigTerm()
        self._registerHandlerForSigKill()





__version__= "$Id:$"



class ServerInfo(object):

    def __init__(self,
                 name,
                 uptimeInSec,
                 hostname,
                 replyPort):
        self.name= name
        self.uptimeInSec= uptimeInSec
        self.hostname= hostname
        self.replyPort= replyPort


    def __repr__(self):
        return "Name: %s, Uptime: %g, Hostname: %s:%s" % (
            self.name,
            self.uptimeInSec,
            self.hostname,
            self.replyPort)



import time
import sys
import zmq
import pickle
from plico.utils.decorator import cacheResult, override, returnsNone
from plico.utils.logger import Logger
from plico.utils.barrier import Barrier, FunctionPredicate, BarrierTimeout
from plico.utils.constants import Constants
from plico.rpc.abstract_remote_procedure_call import \
    AbstractRemoteProcedureCall

if sys.version_info[0] >= 3:
    pickle_options = {'encoding': 'latin1'}
else:
    pickle_options = {}


class ZmqRpcTimeoutError(Exception):
    pass


class ZmqRemoteProcedureCall(AbstractRemoteProcedureCall):

    def __init__(self, timeModule=time):
        self._context = zmq.Context()
        self._logger = Logger.of("ZmqRPC")
        self._timeMod = timeModule

    @override
    @returnsNone
    def publishPickable(self, socket, pickableObject):
        resPickled = pickle.dumps(pickableObject, Constants.PICKLE_PROTOCOL)
        self._logger.debug("sending %s (pickle size: %d)" %
                           (str(pickableObject), len(resPickled)))
        socket.send(resPickled, zmq.NOBLOCK)

    @override
    def receivePickable(self, socket, timeoutInSec=10):
        toBeReturned = self.receiveWithTimeout(socket, timeoutInSec)
        retObj = pickle.loads(toBeReturned, **pickle_options)
        if isinstance(retObj, Exception):
            raise retObj
        else:
            return retObj

    @override
    @returnsNone
    def sendCameraFrame(self, socket, frame):
        self.publishPickable(socket, frame)

    @override
    def recvCameraFrame(self, socket, timeoutInSec=10):
        return self.receivePickable(socket, timeoutInSec)

    def publisherSocket(self,
                        port,
                        connect=False,
                        host='*',
                        hwm=1):
        '''
        Create a PUB-style socket for data publishers.
        If <connect> is true, connects to a XPUB/XSUB forwarding device.
        '''
        socket = self._context.socket(zmq.PUB)
        socket.setsockopt(zmq.SNDHWM, hwm)
        socket.setsockopt(zmq.LINGER, 0)

        try:
            if connect:
                socket.connect(self.tcpAddress(host, port))
            else:
                socket.bind(self.tcpAddress(host, port))
        except Exception as e:
            newMsg = str("%s %s:%d" % (str(e), host, port))
            raise (type(e))(newMsg)
        return socket

    def tcpAddress(self, host, port):
        return "tcp://%s:%d" % (host, port)

    def subscriberSocket(self, host, port, filt=b'', conflate=False):
        '''
        Create a SUB-style socket for data receivers
        '''
        socket = self._context.socket(zmq.SUB)
        if conflate:
            socket.setsockopt(zmq.CONFLATE, 1)

        socket.connect(self.tcpAddress(host, port))
        socket.setsockopt(zmq.SUBSCRIBE, filt)
        return socket

    def xpubsubSockets(self, hostSub, portSub, hostPub, portPub):
        '''
        Creates frontend and backend for a XPUB/XSUB forwarding device
        '''
        frontend_addr = self.tcpAddress(hostSub, portSub)
        backend_addr = self.tcpAddress(hostPub, portPub)

        frontendSocket = self._context.socket(zmq.SUB)
        frontendSocket.bind(frontend_addr)
        frontendSocket.setsockopt(zmq.SUBSCRIBE, b'')

        backendSocket = self._context.socket(zmq.PUB)
        backendSocket.bind(backend_addr)

        return frontendSocket, backendSocket

    @cacheResult
    def replySocket(self, port, host='*'):
        '''
        Create a REP-style socket for servers
        '''
        try:
            socket = self._context.socket(zmq.REP)
            socket.bind(self.tcpAddress(host, port))
        except Exception as e:
            newMsg = str("%s %s:%d" % (str(e), host, port))
            raise (type(e))(newMsg)
        return socket

    @cacheResult
    def requestSocket(self, host, port):
        '''
        Create a REQ-style socket for clients
        '''
        socket = self._context.socket(zmq.REQ)
        socket.connect(self.tcpAddress(host, port))
        return socket

    def _discardPendingAnswers(self, socket):
        try:
            buf = socket.recv(zmq.NOBLOCK)
            self._logger.debug("got pending buffer %d bytes (%s)" %
                               (len(buf), buf))
        except zmq.ZMQError:
            return
        except Exception as e:
            self._logger.debug("got Exception %s" % str(e))

    def _hasSendMultiPartSucceded(self, socket, multiPartMsg):
        try:
            self._discardPendingAnswers(socket)
            socket.send_multipart(multiPartMsg, zmq.NOBLOCK)
            return True
        except zmq.ZMQError as e:
            self._logger.debug("trapped ZMQError on cmd %s: %s" %
                               (multiPartMsg[0].decode(), str(e)))
            return False

    def _sendMultiPartWithBarrierTimeout(self, socket,
                                         multiPartMsg, timeoutSec):
        try:
            Barrier(timeoutSec, 0.1, self._timeMod).waitFor(
                FunctionPredicate.create(self._hasSendMultiPartSucceded,
                                         socket, multiPartMsg))
        except BarrierTimeout as e:
            raise ZmqRpcTimeoutError(str(e))

    def sendRequest(self, socket, cmd, args=(), timeout=10):
        '''
        Perform client request/reply
        Request is a ZMQ multipart message:
         - command string
         - pickled argument list
        Reply is a pickled object
        '''
        self._logger.debug("sending request %s %s" % (cmd, args))
        t0 = time.time()
        self._sendMultiPartWithBarrierTimeout(
            socket,
            [cmd.encode(),
             pickle.dumps(args, Constants.PICKLE_PROTOCOL)],
            timeout)
        toBeReturned = self.receiveWithTimeout(socket, timeout)
        retObj = pickle.loads(toBeReturned, **pickle_options)
        self._logger.debug("%s received %s in %.3fs" % (
            cmd, str(retObj), time.time() - t0))
        if isinstance(retObj, Exception):
            raise retObj
        else:
            return retObj

    def receiveWithTimeout(self, socket, timeoutInSeconds=1):
        poller = zmq.Poller()
        poller.register(socket, zmq.POLLIN)
        msgs = dict(poller.poll(timeoutInSeconds * 1000))
        if socket in msgs and msgs[socket] == zmq.POLLIN:
            return socket.recv(zmq.NOBLOCK)
        else:
            self._logger.debug("raising ZmqRpcTimeoutError")
            raise ZmqRpcTimeoutError()

    @override
    def handleRequest(self, obj, socket, multi=False):
        '''
        Handle one or more requests on a REP socket, with the format
        sent by sendRequest()
        '''
        while 1:
            try:
                msg = socket.recv_multipart(zmq.NOBLOCK)
                method = msg[0].decode()
                try:
                    args = pickle.loads(msg[1], **pickle_options)
                except ValueError as err:
                    self._logger.notice('Request %s failed. Caught %s %s' %
                                        (method, type(err), str(err)))
                    self._sendAnswer(socket, err)
                self._logger.debug("received request %s %s" %
                                   (method, str(args)))
                try:
                    res = getattr(obj, method).__call__(*args)
                except Exception as e:
                    self._logger.notice('Request %s %s failed. Caught %s %s' %
                                        (method, str(args), type(e), str(e)))
                    res = e
                self._sendAnswer(socket, res)
            except zmq.ZMQError:
                return
            except Exception as e:
                self._logger.error("Unknown error %s" % str(e))
            if not multi:
                break

    def _sendAnswer(self, socket, answer):
        resPickled = pickle.dumps(answer, Constants.PICKLE_PROTOCOL)
        self._logger.debug("answering %s (pickle size: %d)" %
                           (str(answer), len(resPickled)))
        socket.send(resPickled, zmq.NOBLOCK)



'''
Base class for dynamic reconnection
'''

class ConnectionException(Exception):
    pass


class ReconnectInfo:

    def __init__(self, connect_func, disconnect_func, exc_to_catch_list):
        self.connect_func = connect_func
        self.disconnect_func = disconnect_func
        self.exceptions_to_catch = (OSError,)
        if exc_to_catch_list:
             self.exceptions_to_catch += tuple(exc_to_catch_list)
        self.connected = False


class Reconnecting:

    def __init__(self, connect_func, disconnect_func, exc_to_catch_list=None):
        '''
        Base class for dynamic reconnection


        connect_func: function with no arguments that will connect to device.
        disconnect_func: function with no arguments that disconnects from device
        exc_to_catch_list: optional sequence of communication exceptions that
                           must be catched and cause a disconnection event.
                           By default only OSError is catched.
        '''
        self._reconnectInfo = ReconnectInfo(connect_func, disconnect_func, exc_to_catch_list)


def reconnect(method):
    '''
    Decorator to make sure that a method is executed
    after connecting to the device, and trigger
    a reconnect in the next command if any error occurs.

    Any communication problem will raise a ConnectionException
    '''
    def wrapped(*args, **kwargs):
        self = args[0]
        info = self._reconnectInfo
        try:
            if not info.connected:
                info.connect_func()
                info.connected = True
            return method(*args, **kwargs)
        except info.exceptions_to_catch as e:
            info.disconnect_func()
            info.connected = False
            raise ConnectionException('Error communicating with device: %s. Will retry...' % str(e))

    return wrapped

# ___oOo___



from plico.utils.loop import Loop
from plico.utils.decorator import override, returns

__version__= "$Id: fake_loop.py 25 2018-01-26 19:00:40Z lbusoni $"


class FakeLoop(Loop):

    def __init__(self):
        Loop.__init__(self)
        self._closed= False
        self._gain= -1.0
        self._convergenceStepCount= 0
        self._converged= False
        self._initialized= False


    @override
    def name(self):
        return "FakeLoop"


    def initialize(self):
        self._initialized= True


    def deinitialize(self):
        self._initialized= False
        self._closed= False
        self._converged= False


    @override
    def performOnePass(self):
        self._convergenceStepCount+= 1


    @override
    def close(self):
        self._convergenceStepCount+= 1
        self._closed= True


    @override
    def open(self):
        self._closed= False


    @override
    @returns(bool)
    def isClosed(self):
        return self._closed


    @returns(bool)
    def isInitialized(self):
        return self._initialized


    @override
    @returns(int)
    def getConvergenceStepCount(self):
        return self._convergenceStepCount


    def setConvergenceFlagTo(self, converged):
        self._converged= converged


    @override
    @returns(bool)
    def hasConverged(self):
        return self._converged



#!/usr/bin/env python

import threading
lock = threading.Lock()


def minimize(*args, **kwargs):
    with lock:
        import lmfit
        return lmfit.minimize(*args, **kwargs)


def leastsq(*args, **kwargs):
    with lock:
        import scipy.optimize
        return scipy.optimize.leastsq(*args, **kwargs)


def medfilt2d(*args, **kwargs):
    with lock:
        import scipy.signal
        return scipy.signal.medfilt2d(*args, **kwargs)


def median_filter(*args, **kwargs):
    with lock:
        import scipy.ndimage.filters
        return scipy.ndimage.filters.median_filter(*args, **kwargs)


def splrep(*args, **kwargs):
    with lock:
        import scipy.interpolate
        return scipy.interpolate.splrep(*args, **kwargs)


def splev(*args, **kwargs):
    with lock:
        import scipy.interpolate
        return scipy.interpolate.splev(*args, **kwargs)


def curve_fit(*args, **kwargs):
    with lock:
        import scipy.optimize
        return scipy.optimize.curve_fit(*args, **kwargs)


def rotate(*args, **kwargs):
    with lock:
        import scipy.ndimage.interpolation
        return scipy.ndimage.interpolation.rotate(*args, **kwargs)


def pinv2(*args, **kwargs):
    with lock:
        import scipy.linalg
        return scipy.linalg.pinv2(*args, **kwargs)


def pinv(*args, **kwargs):
    with lock:
        import scipy.linalg
        return scipy.linalg.pinv(*args, **kwargs)


def gaussian_filter(*args, **kwargs):
    with lock:
        import scipy.ndimage.filters
        return scipy.ndimage.filters.gaussian_filter(*args, **kwargs)


def svdvals(*args, **kwargs):
    with lock:
        import scipy.linalg
        return scipy.linalg.svdvals(*args, **kwargs)


def binom(*args, **kwargs):
    with lock:
        import scipy.special
        return scipy.special.binom(*args, **kwargs)



'''
Serial/USB connection dynamic lookup
'''

import serial.tools.list_ports


class SerialOrUSBException(Exception):
    pass


class SerialOrUSBConnection:

    def __init__(self, comport=None, usb_serial_number=None, usb_vendor_id=None, usb_product_id=None):
        '''
        comport: serial connection port, if known.
        usb_serial_number: device USB serial number, if known. Either the com
                           port or the device serial number must be specified.
        '''
        if (comport is None and usb_serial_number is None and
           ((usb_vendor_id is None) or (usb_product_id is None))):
            raise Exception('At least one of comport, usb_serial_number or usb vendor/product ID must be set')

        self.comport = comport
        self.usb_serial_number = usb_serial_number
        self.usb_vendor_id = usb_vendor_id
        self.usb_product_id = usb_product_id

    @staticmethod
    def fromConfiguration(configuration, section_name):
        try:
            comport = configuration.getValue(section_name, 'port')
        except KeyError:
            comport = None
        try:
            usb_serial_number = configuration.getValue(section_name, 'usb_serial_number')
        except KeyError:
            usb_serial_number = None
        try:
            usb_vendor_id = configuration.getValue(section_name, 'usb_vendor_id')
            usb_vendor_id = int(usb_vendor_id, 0)
        except KeyError:
            usb_vendor_id = None
        try:
            usb_product_id = configuration.getValue(section_name, 'usb_product_id')
            usb_product_id = int(usb_product_id, 0)
        except KeyError:
            usb_product_id = None
        return SerialOrUSBConnection(comport, usb_serial_number, usb_vendor_id, usb_product_id)

    def port_name(self):
        '''
        Returns the serial port name, either because if was specified
        in the constructor, or because it was found by looking up
        the USB serial number. Raises a ConnectionException if not found
        '''
        if self.comport:
            return self.comport

        usbport = self.device_by_serial_number(self.usb_serial_number)
        if usbport:
            return usbport

        usbport = self.device_by_id(self.usb_vendor_id, self.usb_product_id)
        if usbport:
            return usbport

        errmsg = 'No COM port specified and/or USB serial number not found'
        raise SerialOrUSBException(errmsg)

    @staticmethod
    def device_by_serial_number(serial_number):
        '''
        Returns the device path (i.e. '/dev/ttyUSB1') for the
        USB device identified by the serial_number, or None
        if the device is not found.
        '''
        if serial_number is None:
            return None
        ports = serial.tools.list_ports.comports()
        my_ports = [x for x in ports if x.serial_number == serial_number]
        if len(my_ports) > 0:
            return my_ports[0].device
        else:
            return None

    @staticmethod
    def device_by_id(vendor_id, product_id):
        '''
        Returns the device path (i.e. '/dev/ttyUSB1') for the
        USB device identified by the vendor and product id,
        or None if the device is not found.
        '''
        if vendor_id is None or product_id is None:
            return None

        hwid = '%04X:%04X' % (vendor_id, product_id)
        ports = serial.tools.list_ports.comports()
        my_ports = [x for x in ports if hwid in x.hwid]
        if len(my_ports) > 0:
            return my_ports[0].device
        else:
            return None

# ___oOo___




import appdirs
from pkg_resources import Requirement, resource_filename
import shutil
import os
from plico.utils.addtree import mkdirp


class ConfigFileManager():

    def __init__(self, appName, appAuthor, pythonPackageName):
        self._appName = appName
        self._appAuthor = appAuthor
        self._packageName = pythonPackageName
        self._appdirs = appdirs.AppDirs(self._appName, self._appAuthor)

    def getConfigFilePath(self):
        confPath = os.path.join(self._appdirs.user_config_dir,
                                '%s.conf' % self._packageName)
        return confPath

    def _getConfigFilePathInPackage(self):
        return resource_filename(
            Requirement(self._packageName),
            "%s/conf/%s.conf" % (self._packageName, self._packageName))

    def doesConfigFileExists(self):
        return os.path.isfile(self.getConfigFilePath())

    def installConfigFileFromPackage(self, overwrite=False):
        if self.doesConfigFileExists() and (overwrite is False):
            return
        source = self._getConfigFilePathInPackage()
        dest = self.getConfigFilePath()
        mkdirp(os.path.dirname(self.getConfigFilePath()))
        shutil.copyfile(source, dest)



import numpy as np
from math import factorial


class ZernikeGenerator(object):
    """ZernikeGenerator

    A class to generate Zernike polynomials and gradient of Zernike
    polynomials

    Example:
        >>> numberOfPixelOnPupilDiameter= 64
        >>> zg= ZernikeGenerator(numberOfPixelOnPupilDiameter)
        >>> asti= zg.getZernike(5)
        >>> asti[
        >>> type(asti)
        numpy.ma.core.MaskedArray
        >>> asti.shape
        (64,64)
        >>> dTiltdx= zg.getDerivativeX(2)
        >>> dTiltdx.mean()
        2.0
        >>> dTiltdx.ptp()
        0.0

    """

    def __init__(self, nPixelOnDiameter):
        self._nPixel = nPixelOnDiameter
        self._rhoMap, self._thetaMap = self._polar_array(self._nPixel)
        self._dx = None
        self._dy = None
        self._dictCache = {}
        self._dictDxCache = {}
        self._dictDyCache = {}

    def getRadius(self):
        return self._nPixel / 2.

    def _derivativeCoeffX(self, index):
        if (self._dx is None) or (self._dx.shape[0] < index):
            self._dx = self._computeDerivativeCoeffX(index)
        return self._dx[0:index, 0:index]

    def _derivativeCoeffY(self, index):
        if (self._dy is None) or (self._dy.shape[0] < index):
            self._dy = self._computeDerivativeCoeffY(index)
        return self._dy[0:index, 0:index]

    def _degree(self, index):
        n = int(0.5 * (np.sqrt(8 * index - 7) - 3)) + 1
        cn = n * (n + 1) / 2 + 1
        if n % 2 == 0:
            m = int(index - cn + 1) // 2 * 2
        else:
            m = int(index - cn) // 2 * 2 + 1
        radialDegree = n
        azimuthalFrequency = m
        return radialDegree, azimuthalFrequency

    def _rnm(self, radialDegree, azimuthalFrequency, rhoArray):
        n = radialDegree
        m = azimuthalFrequency
        rho = rhoArray
        if (n - m) % 2 != 0:
            raise Exception("n-m must be even. Got %d-%d" % (n, m))
        if abs(m) > n:
            raise Exception("The following must be true |m|<=n. Got %d, %d" %
                            (n, m))
        mask = np.where(rho <= 1, False, True)

        if(n == 0 and m == 0):
            return np.ma.masked_array(data=np.ones(rho.shape), mask=mask)
        rho = np.where(rho < 0, 0, rho)
        Rnm = np.zeros(rho.shape)
        S = (n - abs(m)) // 2
        for s in range(0, S + 1):
            CR = pow(-1, s) * factorial(n - s) / \
                (factorial(int(s)) * factorial(int(-s + (n + abs(m)) / 2)) *
                 factorial(int(-s + (n - abs(m)) / 2)))
            p = CR * pow(rho, n - 2 * s)
            Rnm = Rnm + p
        return np.ma.masked_array(data=Rnm, mask=mask)

    def _polar(self, index, rhoArray, thetaArray):
        n, m = self._degree(index)
        rho = rhoArray
        theta = thetaArray

        Rnm = self._rnm(n, m, rho)
        NC = np.sqrt(2 * (n + 1))
        if m == 0:
            return np.sqrt(0.5) * NC * Rnm
        if index % 2 == 0:
            return NC * Rnm * np.cos(m * theta)
        else:
            return NC * Rnm * np.sin(m * theta)

    def _polar_array(self, nPixel):
        X, Y = np.mgrid[-1 + 1. / nPixel: 1 - 1. / nPixel: nPixel * 1j,
                        -1 + 1. / nPixel: 1 - 1. / nPixel: nPixel * 1j]
        r = np.sqrt(X**2 + Y**2)
        th = np.arccos(np.transpose(X * 1. / r))
        th = np.where(th < 2. * np.pi, th, 0)
        th = np.where(X < 0, 2. * np.pi - th, th)
        return r, th

    def getZernikeDict(self, indexVector):
        ret = {}
        for index in indexVector:
            ret[index] = self.getZernike(index)
        return ret

    def getZernike(self, index):
        """getZernike

        Retrieve a map representing the index-th Zernike polynomial

        Args:
            index (int): The index of Zernike map to be generated,
                following Noll 1976 ordering.

        Returns:
            np.array: A map representing the index-th Zernike polynomial

        """
        if index not in list(self._dictCache.keys()):
            self._dictCache[index] = self._polar(index, self._rhoMap,
                                                 self._thetaMap)
        return self._dictCache[index]

    def __getitem__(self, index):
        return self.getZernike(index)

    def _computeDerivativeCoeffX(self, index):
        jmax = index
        G_mat = np.zeros((jmax, jmax))
        for i in range(1, jmax + 1):
            for j in range(1, jmax + 1):
                ni, mi = self._degree(i)
                nj, mj = self._degree(j)
                if (
                    (
                        (
                            (
                                mi != 0 and mj != 0
                            ) and (
                                (
                                    _isEven(i) and _isEven(j)
                                ) or (
                                    _isOdd(i) and _isOdd(j)
                                )
                            )
                        ) or (
                            (
                                (mi == 0) and _isEven(j)
                            ) or (
                                (mj == 0) and _isEven(i)
                            )
                        )
                    ) and (
                        (mj == mi + 1) or (mj == mi - 1)
                    ) and (
                        j < i
                    )
                ):
                    G_mat[i - 1, j - 1] = np.sqrt((ni + 1) * (nj + 1))
                    if ((mi == 0) or (mj == 0)):
                        G_mat[i - 1, j - 1] *= np.sqrt(2)
        return G_mat

    def _computeDerivativeCoeffY(self, index):
        jmax = index
        G_mat = np.zeros((jmax, jmax))
        for i in range(1, jmax + 1):
            for j in range(1, jmax + 1):
                ni, mi = self._degree(i)
                nj, mj = self._degree(j)
                if (
                    (
                        (
                            (
                                mi != 0 and mj != 0
                            ) and (
                                (
                                    _isOdd(i) and _isEven(j)
                                ) or (
                                    _isEven(i) and _isOdd(j)
                                )
                            )
                        ) or (
                            (
                                (mi == 0) and _isOdd(j)
                            ) or (
                                (mj == 0) and _isOdd(i)
                            )
                        )
                    ) and (
                        (mj == mi + 1) or (mj == mi - 1)
                    ) and (
                        j < i
                    )
                ):
                    G_mat[i - 1, j - 1] = np.sqrt((ni + 1) * (nj + 1))
                    if ((mi == 0) or (mj == 0)):
                        G_mat[i - 1, j - 1] *= np.sqrt(2)
                    if (
                        (
                            (
                                (mj == mi + 1) and _isOdd(i)
                            ) or (
                                (mj == mi - 1) and _isEven(i)
                            )
                        ) and (
                            mi != 0
                        )
                    ):
                        G_mat[i - 1, j - 1] *= -1

        return G_mat

    def getDerivativeXDict(self, indexVector):
        ret = {}
        for index in indexVector:
            ret[index] = self.getDerivativeX(index)
        return ret

    def getDerivativeYDict(self, indexVector):
        ret = {}
        for index in indexVector:
            ret[index] = self.getDerivativeY(index)
        return ret

    def getDerivativeX(self, index):
        if index not in list(self._dictDxCache.keys()):
            self._dictDxCache[index] = self._computeDerivativeX(index)
        return self._dictDxCache[index]

    def _computeDerivativeX(self, index):
        coeffX = self._derivativeCoeffX(index)
        dx = self.getZernike(1) * 0.
        for i in range(1, index):
            dx += coeffX[index - 1, i - 1] * self.getZernike(i)
        return dx

    def getDerivativeY(self, index):
        if index not in list(self._dictDyCache.keys()):
            self._dictDyCache[index] = self._computeDerivativeY(index)
        return self._dictDyCache[index]

    def _computeDerivativeY(self, index):
        coeffY = self._derivativeCoeffY(index)
        dy = self.getZernike(1) * 0.
        for i in range(1, index):
            dy += coeffY[index - 1, i - 1] * self.getZernike(i)
        return dy


def _isOdd(num):
    return num % 2 != 0


def _isEven(num):
    return num % 2 == 0
